import logging
from multiprocessing import Pool
from multiprocessing.dummy import Pool as ThreadPool
import os
import leip


from mad2.util import get_all_mad_files
import mad2.hash

lg = logging.getLogger(__name__)
lg.setLevel(logging.DEBUG)


def calc_madfile_sum(madfile, force=False, echo=False,
                     echo_changed=False):

    lg.warning("DEPRECATED")

    if madfile['filename'] in ['SHA1SUMS', 'QDSUMS']:
        return

    if madfile.get('orphan', False):
        # won't deal with orphaned files
        return

    if madfile.get('isdir', False):
        # won't deal with dirs
        return

    if not (madfile.get('qdhash_changed') or force):
        #probably not changed - ignore
        if echo:
            print(madfile['inputfile'])
            return

    dirname = madfile['dirname']
    filename = madfile['filename']

    lg.debug("creating sha1 for %s", filename)

    sha1file = os.path.join(dirname, 'SHA1SUMS')
    qdhashfile = os.path.join(dirname, 'QDSUMS')

    sha1 = mad2.hash.get_sha1sum(os.path.join(dirname, filename))
    mad2.hash.append_hashfile(sha1file, filename, sha1)

    qd = mad2.hash.get_qdhash(madfile['fullpath'])
    mad2.hash.append_hashfile(qdhashfile, filename, qd)

    if echo or echo_changed:
        print madfile['inputfile']


# @leip.flag('-f', '--force', help='force recalculation')
# @leip.flag('-E', '--echo_changed', help='echo names of recalculated files')
# @leip.flag('-e', '--echo', help='echo all filenames')
# @leip.arg('file', nargs='*')
# @leip.command
# def sha1(app, args):
#     """
#     Echo the filename

#     note - this ensures that the sha1sum is calculated
#     """
#     lg.warning("DEPRECATED")

#     for madfile in get_all_mad_files(app, args):

#         calc_madfile_sum(madfile, args.force, args.echo, args.echo_changed)





from __future__ import print_function

import logging
import os
from pprint import pprint
import re

import leip
import Yaco

from py2neo import neo4j
from py2neo import cypher

neo_logger = logging.getLogger(
    "py2neo.packages.httpstream.http")
neo_logger.setLevel(logging.WARNING)

from mad2.util import get_all_mad_files

lg = logging.getLogger(__name__)
#lg.setLevel(logging.INFO)

class GDB:
    #Dummy object that holds data on a database
    def __init__(self, app):
        self.app = app #leip application - access to conf
        if 'uri' in self.app.conf.plugin.neo4j:
            uri = self.app.conf.plugin.neo4j.uri
            if not uri[-7:] == 'db/data':
                lg.critical("Invalid neo4j uri %s", uri)
                lg.critical(" | should end with /db/data")
                exit(-1)
            lg.debug("Neo4j db: %s", uri)
            self.db = neo4j.GraphDatabaseService(uri)
        else:
            lg.debug("Neo4j db: default")
            self.db = neo4j.GraphDatabaseService()

        self.sha1_index = self.db.get_or_create_index(
                neo4j.Node, "Sha1")
        self.file_index = self.db.get_or_create_index(
                neo4j.Node, "File")
        self.host_index = self.db.get_or_create_index(
                neo4j.Node, "Host")
        self.user_index = self.db.get_or_create_index(
                neo4j.Node, "User")
        self.project_index = self.db.get_or_create_index(
                neo4j.Node, "Project")

DB = None
def get_db(app):
    global DB
    if DB is None:
        DB = GDB(app)
    return DB

def neo_save_madfile(app, madfile):
    db = get_db(app)
    simple = madfile.simple()
    uri = madfile.uri
    user = madfile.username
    host = madfile.host
    sha1 = madfile.hash.sha1
    project = madfile.project

    #del simple['username']
    lg.debug("Saving %s to neo4j", madfile.basename)
    if project:
        del simple['project']

    file_node = db.file_index.get_or_create("uri", uri, simple)
    file_node.delete_properties()

    file_node.set_properties(simple)

    file_node.add_labels('file')

    if project:
        proj_node = db.project_index.get_or_create(
            "project", project, {'project' : project} )
        proj_node.add_labels('project')
        proj_node.get_or_create_path("HAS_FILE", file_node)

    if sha1:
        hash_node = db.sha1_index.get_or_create(
            "sha1", sha1, {'sha1' : sha1} )
        hash_node.add_labels('sha1')
        file_node.get_or_create_path("HAS_SHA1", hash_node)

    host_node = db.host_index.get_or_create("host", host,
        {"host" : host})
    host_node.add_labels('host')
    host_node.get_or_create_path("HAS_FILE", file_node)

    user_node = db.host_index.get_or_create("user", user,
        {"user" : user})
    user_node.add_labels('user')
    user_node.get_or_create_path("OWNS_FILE", file_node)


@leip.command
def neo_clean(app, args):
    db = get_db(app)
    db.db.clear()


@leip.hook("madfile_save", 200)
def neo_hook_save(app, madfile):
    if not app.conf.plugin.neo4j.get('autostore', False):
        return
    lg.debug("start neo4j autosave")
    neo_save_madfile(app, madfile)


@leip.arg('query', help='cypher query to run')
@leip.command
def neo_query(app, args):
    """
    print a single value from a single file
    """
    lg.warning("running query: %s", args.query)
    queries = app.conf.plugin.neo4j.cypher
    query_txt = queries.get(args.query)
    print(query_txt)

    session = cypher.Session(app.conf.plugin.neo4j.uri)
    tx = session.create_transaction()
    tx.append(query_txt)
    res = tx.commit()
    cols = []
    for i, r in enumerate(res[0]):
        if i == 0:
            cols = r.columns
            print("#" + "\t".join(cols))
        print("\t".join([str(r[c]) for c in cols]))


def _dup_report_sum(fileset):
    filesize = fileset[0].filesize
    if not filesize:
        return [], 0
    used_space = 0
    for f in fileset:
        nlink = 1
        if f.nlink: nlink = f.nlink
        used_space += filesize / float(nlink)
    if used_space > filesize:
        to_file = []
        for i, f in enumerate(fileset):
            to_file.append(
                "\t".join([str(f[c]) for c in f.columns])
                + "\n")
        return to_file, used_space - filesize
    else:
        return [], 0

@leip.arg('report_file', help="Report file", nargs='?')
@leip.command
def neo_dup(app, args):
    """
    Summarize duplicate files
    """
    query_txt = app.conf.plugin.neo4j.cypher.duplicates
    lg.debug(query_txt)
    session = cypher.Session(app.conf.plugin.neo4j.uri)
    tx = session.create_transaction()
    tx.append(query_txt)
    res = tx.commit()

    this_sha = None
    fileset = []
    wasted_space = 0
    norec = 0
    repfile = []

    for i, r in enumerate(res[0]):
        norec += 1
        if r.sha1 != this_sha:
            if not this_sha is None:
                tf, ws = _dup_report_sum(fileset)
                wasted_space += ws
                repfile.extend(tf)
            this_sha = r.sha1
            fileset = []
        fileset.append(r)

    if fileset and not this_sha is None:
        tf, ws = _dup_report_sum(fileset)
        wasted_space += ws
        repfile.extend(tf)

    lg.info("retrieved %d records", norec)

    print("Wasted space: {:.0f} Mb".format(wasted_space))

    print(args.report_file)
    if args.report_file:
        with open(args.report_file[0], 'w') as F:
            for line in repfile:
                F.write(line)

@leip.arg('file', nargs='*')
@leip.command
def neo_save(app, args):
    """
    print a single value from a single file
    """
    for madfile in get_all_mad_files(app, args):
        if madfile.orphan is True:
            pass # not doing orphans
        else:
            neo_save_madfile(app, madfile)
            print(madfile.filename)




from __future__ import print_function


import logging
import leip
import os
import re

import Yaco

lg = logging.getLogger(__name__)
#lg.setLevel(logging.DEBUG)

from sqlalchemy.exc import OperationalError
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String

Base = declarative_base()
SESSION = None

from mad2.util import get_all_mad_files

class SqlMadInfo(Base):
    __tablename__ = 'madinfo'
    id = Column(Integer, primary_key=True)
    madfile = Column(Integer, index=True)
    key = Column(String, index=True)
    val = Column(String, index=True)

    def __init__(self, madfileId, key, value):
        self.madfile = madfileId
        self.key = key
        self.val = str(value)

class SqlMadFile(Base):
    __tablename__ = 'madfile'

    id = Column(Integer, primary_key=True)
    sha1 = Column(String, index=True)
    basename = Column(String, index=True)
    dirname = Column(String, index=True)
    host = Column(String, index=True)

    def __init__(self, maf):
        if maf.hash.has_key('sha_1'):
            self.sha1 = maf.hash.sha1
        else:
            self.sha1 = ""
        self.basename = maf.basename
        self.dirname = maf.dirname
        self.host = maf.host

    def __repr__(self):
       return "<SqlMadFile('%s|%s')>" % (self.basename, self.id)

def get_session(app):
    global Base, SESSION
    if SESSION != None:
        return SESSION

    enginestr = app.conf.plugin.sql.engine
    lg.debug("opening db @ {}".format(enginestr))
    engine = create_engine(enginestr)
    SESSION = sessionmaker(bind=engine)()
    Base.metadata.create_all(engine)
    return SESSION

def save_madfile(app, maf):
    lg.debug("saving {}".format(maf.basename))
    session = get_session(app)
    allrecs = session.query(SqlMadFile)\
            .filter(SqlMadFile.basename==maf.basename)\
            .filter(SqlMadFile.dirname==maf.dirname)\
            .filter(SqlMadFile.host==maf.host).all()

    if len(allrecs) > 1:
        lg.error("duplicates in database :(")
        m = allrecs[0]
    elif len(allrecs) == 0:
        m = SqlMadFile(maf)
        session.add(m)
        session.commit()
    else:
        m = allrecs[0]

    #remove all old key/value pairs
    madid = m.id

    session.query(SqlMadInfo)\
            .filter(SqlMadInfo.madfile == madid)\
            .delete()
    session.commit()

    for ky in maf.keys():
        vl = maf.get(ky, None)
        if vl is None:
            continue
        if isinstance(vl, Yaco.Yaco):
            continue

        m = SqlMadInfo(madid, ky, vl)
        session.add(m)
        session.commit()

    return

@leip.arg('value')
@leip.arg('key')
@leip.command
def dbfind(app, args):
    """
    find using the database
    """
    session = get_session(app)
    allrecs = session.query(SqlMadInfo)\
            .filter(SqlMadInfo.key==args.key)\
            .filter(SqlMadInfo.val==args.value).all()
    ids = [x.madfile for x in allrecs]
    mafs = [session.query(SqlMadFile).get(x) for x in ids]
    for m in mafs:
        print(m.host + ':' + os.path.join(m.dirname, m.basename))

@leip.arg('file', nargs='*')
@leip.command
def dbsave(app, args):
    """
    print a single value from a single file
    """
    for madfile in get_all_mad_files(app, args):
        save_madfile(app, madfile)
        print(madfile.filename)

@leip.hook("madfile_save", 200)
def sql_hook_save(app, madfile):
    lg.debug("start save to %s" % app.conf.plugin.sql.engine)
    save_madfile(app, madfile)



"""
Render
"""

import logging
import re

import jinja2



def regex_sub(s, find, replace):
    """A non-optimal implementation of a regex filter"""
    return re.sub(find, replace, s)


def get_jinja_env():

    jenv = jinja2.Environment(
        undefined=jinja2.DebugUndefined)

    jenv.filters['re_sub'] = regex_sub
    return jenv


def recrender(template, data):

    jenv = get_jinja_env()
    rendered = template
    iteration = 0

    if not isinstance(data, list):
        data = [data]

#        data = fanttail.Fanstack
    # stack all data - to prevent potential problems
    # TODO: needs more investigation

    data_stacked = {}
    for d in data[::-1]:
        data_stacked.update(d)

    last = None
    while '{{' in rendered or '{%' in rendered:

        if iteration > 0 and rendered == last:
            # no improvement
            break
        last = rendered

        try:
            template = jenv.from_string(rendered)
        except:
            print("problem creating template with:")
            print(rendered)
            raise

        try:
            rendered = template.render(c=data_stacked, **data_stacked)
        except jinja2.exceptions.UndefinedError:
            pass
        except:
            print("cannot render")
            print(rendered)
            raise

        iteration += 1

    return rendered



from __future__ import print_function

import logging
import os
import socket
import textwrap

import leip

from mad2.util import get_all_mad_files, get_mad_file
from termcolor import cprint

lg = logging.getLogger(__name__)

#
# @leip.arg('-c', '--command_line')
# @leip.arg('-f', '--derived_from', action='append', default=[])
# @leip.arg('-u', '--used', action='append', default=[])
# @leip.arg('-g', '--generated_by', action='append', nargs=3, default=[],
#           help="name version path")


@leip.arg('file', nargs="*")
@leip.command
def raw(app, args):
    """
    Raw dump of data associated with these files
    """
    for madfile in get_all_mad_files(app, args):
        print(madfile.pretty())


@leip.flag('-r', '--raw', help='raw output of provenance data')
@leip.arg('file', nargs="*")
@leip.command
def prov(app, args):
    """
    Show provenance data
    """
    for madfile in get_all_mad_files(app, args):
        if not 'provenance' in madfile:
            # nothing to show - continue
            continue

        prov_data = madfile['provenance']
        prov_keys = sorted(prov_data.keys())
        latest_key = prov_keys[-1]
        prov = prov_data[latest_key]

        if args.raw:
            print("provenance_key: {}".format(latest_key))
            print(prov.pretty())
            return

        def ccp(*args, **kwargs):
            if not 'end' in kwargs:
                kwargs['end'] = ''
            cprint(*args, **kwargs)

        def cckv(key, val, **kwargs):
            cprint(key, "yellow", end=": ")
            cprint(val)

        # pretty output
        cckv("Date", prov['stopped_at_time'])
        cckv("Tool", prov['tool_name'])
        version = prov['tool_version']
        if len(version) > 50:
            ccp('Version: ', "yellow")
            for i, line in enumerate(textwrap.wrap(version,
                                                   initial_indent="         ",
                                                   subsequent_indent="     ")):
                if i == 0:
                    line = line.strip()
                print(line)
        else:
            cckv("Version", prov['tool_version'])

        ccp("Command line:", "yellow", end="\n")
        print(" \\\n".join(textwrap.wrap(prov['kea_command_line'],
                                         initial_indent='  ',
                                         subsequent_indent='       ')))
        cprint("Related files:", "yellow")
        this_host = socket.gethostname()

        for filename in prov['derived_from']:
            finf = prov['derived_from'][filename]
            ccp("  " + finf['category'], 'magenta')
            ccp("/" + filename, "blue")
            ccp("\n")
            ccp("    Host: ", "yellow")
            if finf['host'] == this_host:
                ccp("{host}\n".format(**finf), "green")
            else:
                ccp("{host}\n".format(**finf), "red")

            ccp("    Path: ", "yellow")
            if finf['host'] == this_host:
                if os.path.exists(finf['filename']):
                    ccp("{filename}\n".format(**finf), "green")
                else:
                    ccp("{filename}\n".format(**finf), "red")
            else:
                ccp("{filename}\n".format(**finf), "grey")


            ccp("    Sha1sum: ", "yellow")
            if finf['host'] == this_host:
                fmaf = get_mad_file(app, finf['filename'])
                if fmaf['sha1sum'] == finf['sha1sum']:
                    ccp("{sha1sum}\n".format(**finf), "green")
                else:
                    ccp("{sha1sum}\n".format(**finf), "red")
            else:
                ccp("{sha1sum}\n".format(**finf), "grey")

#        print(finf.pretty())#





import arrow
import os
import hashlib
import logging
import uuid

from lockfile import FileLock

import mad2.util

lg = logging.getLogger(__name__)
#lg.setLevel(logging.DEBUG)

SHA1CACHE = {}


def get_sha1sum_mad(madfile):
    """
    One function to rule them all!
    """

    global SHA1CACHE

    if madfile['orphan'] == True:
        lg.debug("ORPHAN ALERT!")
        madfile.mad['sha1sum'] = ""
#        madfile.all['sha1sum'] = ""
        return

    lg.debug("get_sha1sum_mad starting for %s", madfile['inputfile'])
    now_time = str(arrow.get(madfile['mtime']).to('local'))
    now_size = madfile['filesize']

    dirname = madfile['dirname']
    filename = madfile['filename']
    fullpath = madfile['fullpath']

    sha1file = os.path.join(dirname, 'SHA1SUMS')
    qdfile = os.path.join(dirname, 'QDSUMS')

    metafile = os.path.join(dirname, 'SHA1SUMS.META')

    # first check metafile
    meta_time, meta_size = None, None
    if os.path.exists(metafile):
        with open(metafile) as F:
            for line in F:
                meta_time, meta_size, meta_file = line.strip().split(None, 2)
                if meta_file == filename:
                    break
            else:
                meta_time, meta_size = None, None

    stored_sha1 = check_hashfile(sha1file, filename)

    #print(now_time, meta_time)
    lg.debug("stored sha1: %s", stored_sha1)
    # lg.debug(" * now time  : %s", now_time)
    # lg.debug(" * meta time : %s", meta_time)
    # lg.debug(" * now size  : %s", now_size)
    # lg.debug(" * meta size : %s", meta_size)

    if (not stored_sha1 is None) \
            and now_time == meta_time \
            and now_size == int(meta_size):

        # all is well - metadata has not changed - return sha1
        lg.debug("metadata has not changed - use sha1sum from SHA1SUMS")
        madfile.mad['sha1sum'] = stored_sha1
        return

    #print (meta_time, meta_size)

    if  (not stored_sha1 is None) \
            and  (meta_time is None) \
            and (meta_size is None):

        # no metadata - check if there is a qd hash
        # For the time being - see if the QDHASH is correct - until all
        # hash files have a meta file
        file_qd = check_hashfile(qdfile, filename)
        if file_qd:
            now_qd = get_qdhash(fullpath)
            if now_qd == file_qd:

                # assume all is well.. (re-)store the sha1
                # so that the metadata gets stored as well.
                #sha1 = check_hashfile(sha1file, filename)

                new_store_sha1(sha1file, metafile, filename,
                               stored_sha1, now_time, now_size)
                madfile.all['sha1sum'] = stored_sha1
                return

    # we need to (re-)calculate the SHA1SUM
    #lg.warning("C:" + filename)
    if stored_sha1 is None:
        # no sha1 - assuming this is the first time
        lg.info("Calculating SHA1SUM for %s", madfile['inputfile'])
    else:
        # there is one, but metadata/qdsum does not match - recalc
        lg.info("Recalculating sha1sum for %s", madfile['inputfile'])
        lg.info("Old sha1sum: %s", stored_sha1)

    sha1 = get_sha1sum(fullpath)
    lg.debug("Current sha1sum: %s", sha1)

    # and save (regardless - if only to update the metadata)
    new_store_sha1(sha1file, metafile, filename,
                   sha1, now_time, now_size)

    madfile.all['sha1sum'] = sha1 #should not be necessary.
    madfile.mad['sha1sum'] = sha1

    if not stored_sha1 is None and \
            sha1 != stored_sha1:
        lg.warning("File changed: %s", madfile['inputfile'])

        #load the old data!
        lg.warning("load data from old record (%s)", stored_sha1)
        madfile.load(stored_sha1)

        # store old sha1 for posterity
        print madfile.mad['old_sha1sum']
        if not isinstance(madfile.mad['old_sha1sum'], list):
            madfile.mad['old_sha1sum'] = []
        while stored_sha1 in madfile.mad['old_sha1sum']:
            madfile.mad['old_sha1sum'].remove(stored_sha1)
        madfile.mad['old_sha1sum'].insert(0, stored_sha1)


        #since the sha1 is stored in the database - reset this
        #to make sure we use the new sha1sum from now on
        madfile.all['sha1sum'] = sha1 #should not be necessary.
        madfile.mad['sha1sum'] = sha1

        madfile.save()


def new_store_sha1(sha1file, metafile, filename,
                   sha1, atime, size):

    hashes = {}
    metas = {}

    # TODO: Figure out if I an do this with file locks - so far it keeps
    #       on giving problems - os for now I'm not.
    #       Not using a lock may result in losing the SHA1SUM - but that
    #       should only result in time loss - since data is linked against
    #       the actual sha1sum.
    lg.debug("writing hash to %s", sha1file)
    # with FileLock(sha1file):
     # read old sha1file
    if os.path.exists(sha1file):
        with open(sha1file) as F:
            for line in F:
                _hsh, _fn = line.strip().split(None, 1)
                hashes[_fn] = _hsh

    # read old metafile
    if os.path.exists(metafile):
        with open(metafile) as F:
            for line in F:
                _date, _size, _fn = line.strip().split(None, 2)
                metas[_fn] = (_date, _size)

    # insert our sha1 - possibly overwriting other version
    hashes[filename] = sha1

    # insert our metadata - possible overwriting previous data
    metas[filename] = (atime, size)

    # write new sha1file
    with open(sha1file, 'w') as F:
        for fn in sorted(hashes.keys()):
            if fn in ['QDSUMS', 'SHA1SUMS']:
                continue
            F.write("{}  {}\n".format(hashes[fn], fn))

    # write new metafile
    with open(metafile, 'w') as F:
        for fn in sorted(metas.keys()):
            if fn in ['QDSUMS', 'SHA1SUMS']:
                continue
            _date, _size = metas[fn]
            #print ('meta', fn, _date, _size)
            F.write("{}  {}  {}\n".format(_date, _size, fn))

#    lg.debug("finished writing hash to %s", sha1file)

#    sha1lock.release()


def append_hashfile(hashfile, filename, hash, date, size):
    lg.warning("DEPRECATED")
    hashes = {}
    metas = {}

    metafile = hashfile + '.META'

    with FileLock(hashfile):
        with FileLock(metafile):
            # read old sha1file
            if os.path.exists(hashfile):
                with open(hashfile) as F:
                    for line in F:
                        hsh, fn = line.strip().split(None, 1)
                        hashes[fn] = hsh

            # read old metafile
            if os.path.exists(metafile):
                with open(metafile) as F:
                    for line in F:
                        date, size, fn = line.strip().split(None, 2)
                        hashes[fn] = (date, size)

            # insert our sha1 - possibly overwriting other version
            hashes[filename] = hash
            # dates[filename =]

            # write new sha1file
            with open(hashfile, 'w') as F:
                for fn in sorted(hashes.keys()):
                    if fn in ['QDSUMS', 'SHA1SUMS']:
                        continue
                    F.write("{}  {}\n".format(hashes[fn], fn))


def get_or_create_sha1sum(filename):
    """
    Get a sha1sum, if it does not exist.

    Also, if there is a qdsum, and it has changed - force
    recalculation of the sha1sum

    """
    lg.warning("DEPRECATED")
    # import traceback
    # traceback.print_stack()

    dirname, basename = os.path.split(filename)
    sha1file = os.path.join(dirname, 'SHA1SUMS')
    qdsumfile = os.path.join(dirname, 'QDSUMS')

    sha1 = check_hashfile(sha1file, basename)
    qd_old = check_hashfile(qdsumfile, basename)
    qd_now = get_qdhash(filename)

    if (sha1 is None) or (qd_old != qd_now):
        sha1 = get_sha1sum(filename)
        append_hashfile(sha1file, basename, sha1)
        append_hashfile(qdsumfile, basename, qd_now)
    return sha1


HASH_CACHE = {}


def check_hashfile(hashfile, filename):
    """
    Check a hashfile & return the checksum
    """

    global HASH_CACHE

    if not os.path.exists(hashfile):
        return None

    # cached??
    hckey = (hashfile, filename)
    if hckey in HASH_CACHE:
        return HASH_CACHE[hckey]

    with open(hashfile) as F:
        for line in F:
            hsh, fn = line.strip().split(None, 1)
            HASH_CACHE[(hashfile, fn)] = hsh

    return HASH_CACHE.get(hckey)


def get_sha1sum(filename):
    """
    Calculate the sha1sum

    """

    if not os.path.exists(filename):
        # not sure what to do with files that do not exist (yet)
        return None

    h = hashlib.sha1()

    blocksize = 2 ** 20
    with open(filename, 'rb') as F:
        for chunk in iter(lambda: F.read(blocksize), b''):
            h.update(chunk)
    return h.hexdigest()


@mad2.util.memoized
def get_qdhash(filename):
    """
    Provde a quick & dirty hash -

    by no means secure - but if used with care, then I'm
    reasaonbly sure that the chance of a collision is small.

    .. and it is fairly fast

    """

#    lg.critical("MM")
    if not os.path.exists(filename):
        # not sure what to do with files that do not exist (yet)
        return None

    if os.path.isdir(filename):
        return None


    sha1sum = hashlib.sha1()
    filesize = os.stat(filename).st_size
    if filesize < 20000:
        with open(filename, 'rb') as F:
            sha1sum.update(F.read())
    else:
        with open(filename, 'rb') as F:
            for x in range(9):
                F.seek(int(filesize * (x / 10.0)))
                sha1sum.update(F.read(2000))

            F.seek(-2000, 2)
            sha1sum.update(F.read())

#    return sha1sum.hexdigest()[:24]#
    return sha1sum.hexdigest()



import logging
import os

import fantail

from mad2.exception import MadPermissionDenied
from mad2.recrender import recrender

lg = logging.getLogger(__name__)
# lg.setLevel(logging.DEBUG)


def dummy_hook_method(*args, **kw):
    return None


STORES = None


class MadFile(fantail.Fanstack):

    """
    Represents a single file
    """

    def __init__(self,
                 inputfile,
                 stores=None,
                 base=fantail.Fantail(),
                 hook_method=dummy_hook_method):

        self.stores = stores
        self.hook_method = hook_method

        lg.debug('madfile start %s', inputfile)
        super(MadFile, self).__init__(
            stack=[fantail.Fantail(),
                   base.copy()])

        self.dirmode = False
        if os.path.isdir(inputfile):
            self.dirmode = True
            dirname = inputfile
            filename = ''
        else:
            dirname = os.path.dirname(inputfile)
            filename = os.path.basename(inputfile)

        lg.debug(
            "Instantiating a madfile for '{}' / '{}'".format(
                dirname, filename))

        self.all['inputfile'] = inputfile
        self.all['dirname'] = os.path.abspath(dirname)
        self.all['filename'] = filename
        self.all['fullpath'] = os.path.abspath(inputfile)

        self.mad['sha1sum'] = ""
        #self.all['sha1sum'] = ""

        if not os.path.exists(inputfile):
            self.all['orphan'] = True

        for s in self.stores:
            store = self.stores[s]
            store.prepare(self)

        self.load()

    def render(self, template, data):
        """
        Render a template from, adding self to the context
        """
        if not isinstance(data, list):
            data = [data]
        return recrender(template, [self] + data)

    @property
    def mad(self):
        return self.stack[0]

    @property
    def all(self):
        return self.stack[1]

    def __str__(self):
        return '<mad2.madfile.MadFile {}>'.format(self['inputfile'])

    def check_sha1sum(self):
        import mad2.hash
        mad2.hash.get_sha1sum_mad(self)

    def on_change(self):
        #call when the file has been changed
        pass

    def on_delete(self):
        #call when the file has been is deleted
        pass

    def load(self, sha1sum=None):
        """
        load the record from the database, possibly with an
        alternative id (used when files change)
        """

        self.all['orphan'] = False \
            if os.path.exists(self.all['inputfile']) \
            else True

        self.hook_method('madfile_pre_load', self)

        fis = self.get('filesize', -1)
        if fis < 1:
            lg.info("file size zero, skip: {}".format(self['inputfile']))
        else:
            for s in self.stores:
                store = self.stores[s]
                store.load(self, sha1sum=sha1sum)

        self.hook_method('madfile_load', self)
        self.hook_method('madfile_post_load', self)

    def save(self):
        self.hook_method('madfile_save', self)
        self.hook_method('madfile_pre_save', self)

        fis = self.get('filesize', -1)
        if fis <= 0:
            lg.info("file size zero - not annotating this")
        else:
            for s in self.stores:
                store = self.stores[s]
                store.save(self)

        self.hook_method('madfile_post_save', self)



class MadDummy(MadFile):
     def __init__(self, data_core, data_all,
                 stores=None,
                 hook_method=dummy_hook_method):

        self.stores = stores
        self.hook_method = hook_method

        super(MadFile, self).__init__(
            stack=[data_core, data_all])

        lg.debug("Instantiating a dummy madfile")

        for s in self.stores:
            store = self.stores[s]
            store.prepare(self)




import logging
import os
import sys
import tempfile
import subprocess

from dateutil.parser import parse as dateparse
import leip
import fantail
from termcolor import cprint

from mad2.util import  get_mad_file, get_all_mad_files, get_filenames
from mad2.ui import message, error, errorexit
import mad2.ui

lg = logging.getLogger(__name__)


##
## define unset
##
@leip.arg('file', nargs="+")
@leip.arg('key')
@leip.arg('-e', '--echo', help='echo ')
@leip.command
def unset(app, args):
    """
    Remove a key/value from a file
    """
    lg.debug("unsetting: %s".format(args.key))
    key = args.key
    keyinfo = app.conf['keywords.{}'.format(key)]
    if keyinfo.get('cardinality', '1') == '+':
        errorexit("Not implemented - unsetting keys with cardinality > 1")

    for madfile in get_all_mad_files(app, args):
        #print(madfile)
        #print(madfile.mad.pretty())
        if args.key in madfile.mad:
            del(madfile.mad[args.key])

        madfile.save()

def _getkeyval(app, key, val, force):

    # First check if this will become a list.
    list_mode = False
    if key[0] == '+':
        lg.debug("treating {} as a list".format(key))
        list_mode = True
        key = key[1:]

    keyinfo = app.conf['keywords.{}'.format(key)]
    keytype = keyinfo.get('type', 'str')

    if (not force) and (not keyinfo.get('description')):
        errorexit('invalid key: "{0}" (use -f?)'.format(key))

    if list_mode and str(keyinfo.get('cardinality', '1')) == '1':
        errorexit("Cardinality == 1 - no lists!")
    elif keyinfo.get('cardinality', '1') == '+':
        list_mode = True

    if keytype == 'int':
        try:
            val = int(val)
        except ValueError:
            lg.error("Invalid integer: %s" % val)
            sys.exit(-1)
    elif keytype == 'float':
        try:
            val = float(val)
        except ValueError:
            lg.error("Invalid float: %s" % val)
            sys.exit(-1)
    elif keytype == 'boolean':
        if val.lower() in ['1', 'true', 't', 'yes', 'aye', 'y', 'yep']:
            val = True
        elif val.lower() in ['0', 'false', 'f', 'no', 'n', 'nope']:
            val = False
        else:
            lg.error("Invalid boolean: %s" % val)
            sys.exit(-1)
    elif keytype == 'date':
        try:
            val = dateparse(val)
        except ValueError:
            lg.error("Invalid date: %s" % val)
            sys.exit(-1)
        lg.debug("date interpreted as: %s" % val)

    if keytype == 'restricted':
        allowed = keyinfo['allowed']
        if not val in allowed.keys():
            errorexit("Value '{0}' not allowed for key '{1}'".format(val, key))

    return key, val, list_mode

@leip.arg('file', nargs='*')
@leip.arg('-k', '--kv', help='key & value to set', metavar=('key', 'val'),
                nargs=2, action='append')
@leip.usage("usage: mad mset [-h] [-f] [-e] -k key val [[-k key val] ...] [file [file ...]]")
@leip.arg('-f', '--force', action='store_true', help='apply force')
@leip.arg('-e', '--echo', action='store_true', help='echo filename')
@leip.command
def mset(app, args):
    """
    Set multiple key/value pairs.
    """
    all_kvs = []
    for k, v in args.kv:
        all_kvs.append(_getkeyval(app, k, v, args.force))

    for madfile in get_all_mad_files(app, args):
        print madfile
        for key, val, list_mode in all_kvs:
            if list_mode:
                if not key in madfile:
                    oldval = []
                else:
                    oldval = madfile[key]
                    if not isinstance(oldval, list):
                        oldval = [oldval]
                madfile.mad[key] = oldval + [val]

            else:
                #not listmode
                madfile.mad[key] = val

        if args.echo:
            print(madfile.filename)
        madfile.save()


@leip.flag('-t', '--transient', help='show transient keywords')
@leip.command
def keywords(app, args):
    """
    Show allowed keywords
    """
    maxkeylen = 0

    for key, keyinfo in app.conf['keywords'].iteritems():
        if keyinfo['hide']:
            continue
        maxkeylen = max(maxkeylen, len(key))

    for key in sorted(app.conf['keywords'].keys()):
        keyinfo = app.conf['keywords'][key]
        transient = keyinfo['transient']
        if keyinfo['hide']:
            continue

        if transient and not args.transient:
            continue

        cprint(('{:' + str(maxkeylen) + '}').format(key), 'yellow',
               end=' : ')

        cprint(str(keyinfo['description']), end="")
        if transient:
            cprint(" (transient)", 'grey')
        else:
            print
        if keyinfo['type'] == 'restricted':
            for i, allowed in enumerate(keyinfo['allowed']):
                ad = keyinfo['allowed'][allowed]
                if i == 0:
                    cprint(('  allowed: -'), end='')
                else:
                    cprint(('           -'), end='')
                cprint(('{}').format(allowed), "green", end=': ')
                cprint(ad)



@leip.arg('-d', '--editor', action='store_true', help='open an editor')
@leip.arg('-f', '--force', action='store_true', help='apply force')
@leip.arg('-p', '--prompt', action='store_true', help='show a prompt')
@leip.arg('file', metavar='dir', nargs='?', default='.')
@leip.arg('value', help='value to set')
@leip.arg('key', help='key to set')
@leip.command
def dset(app, args):
    """
    Like set, but at the directory level
    """
    args.dir = True
    args.echo = False
    madset(app, args)


@leip.arg('-f', '--force', action='store_true', help='apply force')
@leip.arg('-p', '--prompt', action='store_true', help='show a prompt')
@leip.arg('-D', '--dir', action='store_true', help='dir modus - write ' +
          'to ./mad.config files to be read for all files in the dir ' +
          '(and below)')
@leip.arg('-d', '--editor', action='store_true', help='open an editor')
@leip.arg('-e', '--echo', action='store_true', help='echo filename')
@leip.arg('file', nargs='*')
@leip.arg('value', help='value to set', nargs='?')
@leip.arg('key', help='key to set')
@leip.commandName('set')
def madset(app, args):
    """
    Set a key/value for one or more files.

    Use this command to set a key value pair for one or more files.

    This command can take the following forms::

        mad set project test genome.fasta
        ls *.fasta | mad set project test
        find . -size +10k | mad set project test

    """

    key = args.key
    val = args.value


    if args.prompt or args.editor:
        if not args.value is None:
            # when asking for a prompt - the next item on sys.argv
            # is assumed to be a file, and needs to be pushed
            # into args.file
            args.file = [args.value] + args.file

    madfiles = []

    #gather all madfiles for later parsing
    use_stdin = not (args.prompt or args.editor)
    if args.dir:
        for m in get_filenames(args, use_stdin, allow_dirs=True):
            if not os.path.isdir(m):
                continue

            fn = os.path.join(m, 'mad.config')
            mf = fantail.Fantail()
            if os.path.exists(fn):
                mf = fantail.yaml_file_loader(fn)
            mf._mad_dir_name = m
            mf._mad_file_name = fn
            madfiles.append(mf)
    else:
        for m in get_all_mad_files(app, args, use_stdin):
            madfiles.append(m)

    #check if mad needs to show a prompt or editor
    if val is None and not (args.prompt or args.editor):
        args.prompt = True

    # show prompt or editor
    if args.prompt or args.editor:
        # get a value from the user

        default = ''
        #Show a prompt asking for a value
        data = madfiles[0]
        default = madfiles[0].get(key, "")

        if args.prompt:
            sys.stdin = open('/dev/tty')
            val = mad2.ui.askUser(key, default, data)
            sys.stdin = sys.__stdin__

        elif args.editor:
            editor = os.environ.get('EDITOR','vim')
            tmp_file = tempfile.NamedTemporaryFile('wb', delete=False)

            #write default value to the tmp file
            if default:
                tmp_file.write(default + "\n")
            else:
                tmp_file.write("\n")
            tmp_file.close()

            tty = open('/dev/tty')

            subprocess.call('{} {}'.format(editor, tmp_file.name),
                stdin=tty, shell=True)
            sys.stdin = sys.__stdin__


            #read value back in
            with open(tmp_file.name, 'r') as F:
                #removing trailing space
                val = F.read().rstrip()
            #remove tmp file
            os.unlink(tmp_file.name)

    #process key & val
    key, val, list_mode = _getkeyval(app, key, val, args.force)

    # Now process madfiles
    lg.debug("processing %d files" % len(madfiles))


    for madfile in madfiles:
        if list_mode:
            if not key in madfile:
                oldval = []
            else:
                oldval = madfile.get(key)
                if not isinstance(oldval, list):
                    oldval = [oldval]
            if args.dir:
                madfile[key] = oldval + [val]
            else:
                madfile.mad[key] = oldval + [val]
        else:
            #not listmode
            if args.dir:
                madfile[key] = val
            else:
                madfile.mad[key] = val

        if args.echo:
            if args.dir:
                print(madfile._mad_dir_name)
            else:
                print(madfile['filename'])

        if args.dir:
            fantail.yaml_file_save(madfile, madfile._mad_file_name)
        else:
            madfile.save()





import os
import hashlib
import uuid


def get_qdhash(filename):
    """
    Provde a quick & dirty hash -

    by no means secure - but a chance of a collision is small.

    .. and this is reasonably fast.
    """
    lg.warning("DEPRECATED")
    if os.path.isdir(filename):

        #qdid for directories is a uuid - stored in .mad/qid
        maddir = os.path.join(filename, '.mad')
        if not os.path.exists(maddir):
            os.makedirs(maddir)
        qidfile = os.path.join(maddir, 'qid')
        if not os.path.exists(qidfile):
            u = str(uuid.uuid4()).replace('-', '')[:24]
            with open(qidfile, 'w') as F:
                F.write(u)
            return u
        else:
            with open(qidfile) as F:
                u = F.read().strip()
            return u

    sha1sum = hashlib.sha256()
    filesize = os.stat(filename).st_size
    if filesize < 20000:
        with open(filename, 'rb') as F:
            sha1sum.update(F.read())
    else:
        with open(filename, 'rb') as F:
            for x in range(9):
                F.seek(int(filesize * (x / 10.0)))
                sha1sum.update(F.read(2000))

            F.seek(-2000, 2)
            sha1sum.update(F.read())

    return sha1sum.hexdigest()[:24]



# -*- coding: utf-8 -*-
from __future__ import print_function,  unicode_literals

import logging
from datetime import datetime
from signal import signal, SIGPIPE, SIG_DFL
import os
import sys

os.environ['TERM'] = 'xterm' #prevent weird output excape code

import arrow
import leip
from termcolor import colored

import mad2.ui
import mad2.util

# Ignore SIG_PIPE and don't throw exceptions
# otherwise it crashes when you pipe into, for example, head
# see http://tinyurl.com/nwpwyoj
# see http://docs.python.org/library/signal.html
signal(SIGPIPE, SIG_DFL)


lg = logging.getLogger(__name__)


PROFILE = False

def dispatch():
    """
    Run the app - this is the actual application entry point
    """
    if PROFILE:
        import cProfile
        import os
        import pstats
        import tempfile

        pr = cProfile.Profile()
        pr.enable()
        app.run()
        pr.disable()
        handle = tempfile.NamedTemporaryFile(
            delete=False, dir=os.getcwd(), prefix='Mad2.', suffix='.profiler')
        sortby = 'tottime'
        ps = pstats.Stats(pr, stream=handle).sort_stats(sortby)
        ps.print_stats()
        handle.close()
    else:
        app.run()


#
# define show
#
@leip.arg('file', nargs='*')
@leip.flag('-r', '--raw', help='output yaml representation')
@leip.flag('-t', '--tsv', help='output tab delimited representation '
          + '(top level only)')
@leip.command
def show(app, args):
    """ Show mad annotation of one or more file(s)

    default shows a screen formatted output
    """


    i = 0

    if args.tsv:
        for madfile in mad2.util.get_all_mad_files(app, args):
            if i > 0:
                print('---')

            for k in sorted(madfile.keys()):
                v = madfile[k]

                v = _format_value(madfile[k])

            i += 1
    elif args.raw:
        for madfile in mad2.util.get_all_mad_files(app, args):
            x = madfile.all.copy()
            x.update(madfile.mad)
            print(x.pretty())

    else:
        for madfile in mad2.util.get_all_mad_files(app, args):
            print_nicely(app, madfile)

def _format_value(v):
    if isinstance(v, dict):
        return '...'
    elif isinstance(v, list):
        return v[:5]
    elif isinstance(v, datetime):
        return str(arrow.get(v).to('local'))
    else:
        return str(v)

def print_nicely(app, madfile):
    implicit = []
    explicit = []
    max_key_len = 0
    for k in madfile.keys():
        max_key_len = max(len(k), max_key_len)
        v = _format_value(madfile[k])
        if k in madfile.mad:
            explicit.append((k,v))
        else:
            implicit.append((k,v))

    fs = '{} {}: {}'
    ffs = '%-' + str(max_key_len) + 's'

    for k, v in sorted(explicit):
        if k in ['_id', '_id_dump']:
            continue

        print(fs.format(
              colored('x', 'red'),
              ffs % k,
              v))

    for k, v in sorted(implicit):
        print(fs.format(
              colored('i', 'cyan'),
              ffs % k,
              v))


@leip.commandName('key')
def madkeylist(app, args):
    kyw = sorted(app.conf.keywords)
    mxlen = str(max([len(x) for x in kyw]) + 1)
    for c in sorted(app.conf.keywords):
        print(("{:<" + mxlen + "} {}")
              .format(c, app.conf.keywords[c].description))


@leip.commandName('category')
def madcatlist(app, args):
    kyw = sorted(app.conf.keywords.category.allowed)
    mxlen = str(max([len(x) for x in kyw]) + 1)
    for c in sorted(kyw):
        dsc = app.conf.keywords.category.allowed[c]
        print(("{:<" + mxlen + "} {}").format(c, dsc))
        tmpl = app.conf.template[c]
        if not tmpl:
            continue
        mxlen2 = str(max([len(x) for x in tmpl]) + 1)
        for k in sorted(tmpl):
            print(("  - {:<" + mxlen2 + "} {}")
                  .format(k, app.conf.template[c][k]))


@leip.arg('section', help='section to print', default='', nargs='?')
@leip.command
def sysconf(app, args):
    c = app.conf[args.section]
    print(c.pretty())


@leip.arg('comm', metavar='command', help='command to check')
@leip.command
def has_command(app, args):
    comm = args.comm
    lg.debug("checking if we know command %s" % comm)
    if comm in app.leip_commands:
        sys.exit(0)
    else:
        sys.exit(-1)


#
# Instantiate the app and discover hooks & commands
#
#logging.getLogger('leip').setLevel(logging.DEBUG)

app = leip.app('mad2')

# app.parser.add_argument('--trust', action='store_true', help='trust sha1sum')
# discover hooks in this module!
app.discover(globals())



from __future__ import print_function

import collections
import logging
import sys

import fantail
import leip

from mad2.util import get_all_mad_files

lg = logging.getLogger(__name__)

def _getter(o, k):
    assert(isinstance(o, fantail.Fantail))
    if '.' in k:
        k1, k2 = k.split('.', 1)
        return _getter(o[k1], k2)
    else:
        return o[k]


@leip.arg('-s', '--sep',  help='separator', default='\t')
@leip.arg('key', help='keys to display', nargs='+')
@leip.command
def table(app, args):
    """
    Create a table
    """
    print('#' + args.sep.join(map(str, args.key)))
    for madfile in get_all_mad_files(app, args):
        values = []
        for k in args.key:
            vv = _getter(madfile.mad, k)
            if str(vv) == '{}':
                vv = _getter(madfile.all, k)
            if str(vv) == '{}':
                vv = ""
            values.append(vv)

        print(args.sep.join(map(str, values)))


@leip.arg('-s', '--sep',  help='separator', default='\t')
@leip.arg('formula', help='groups = numeric variables', nargs='+')
@leip.command
def groupby(app, args):
    """
    Create a table
    """
    #print(args.formula)
    if not '-' in args.formula:
        print("Need to provide a formula, for example:")
        print("mad groupby project username - filesize")
        sys.exit(-1)

    groups = args.formula[:args.formula.index('-')]
    numers = args.formula[args.formula.index('-')+1:]

    print("#" + args.sep.join(groups + numers))
    def empty_numer():
        return [0] * len(numers)

    data = collections.defaultdict(empty_numer)

    for madfile in get_all_mad_files(app, args):
        gv = []
        nm = []

        for g in groups:
            vv = _getter(madfile.mad, g)
            if str(vv) == '{}':
                vv = _getter(madfile.all, g)
            if str(vv) == '{}':
                vv = 'n.d.'
            gv.append(vv)

        for n in numers:
            vv = _getter(madfile.mad, n)
            if str(vv) == '{}':
                vv = _getter(madfile.all, n)

            if str(vv) == '{}':
                vv = 0
            else: vv = int(vv)
            nm.append(vv)

        for i, n in enumerate(nm):
            data[tuple(gv)][i] += n

        #print(gv, nm, data[tuple(gv)][0] / 1000.)

    for k in sorted(data.keys()):
        rv = map(str, list(k))
        rv.extend(map(str, data[k]))
        print(args.sep.join(rv))




from __future__ import print_function

import logging

import leip
from mad2.util import get_all_mad_files

lg = logging.getLogger(__name__)


@leip.arg('file', nargs='*')
@leip.command
def backup(app, args):
    """
    Shortcut to mark a file for backup
    """
    for madfile in get_all_mad_files(app, args):
    	madfile.mad.backup = True
        madfile.save()





import argparse
import collections
import logging
from multiprocessing import Pool, Lock, Manager
from multiprocessing.dummy import Pool as ThreadPool
import os
import sys
import time

import yaml


from mad2 import hash
from mad2 import util

from colorlog import ColoredFormatter
color_formatter = ColoredFormatter(
    "%(log_color)s%(name)s:%(reset)s "+
    "%(blue)s%(message)s %(purple)s(%(threadName)s)",
    datefmt=None, reset=True,
    log_colors={'DEBUG':    'cyan',
                'INFO':     'green',
                'WARNING':  'yellow',
                'ERROR':    'red',
               'CRITICAL': 'red'})

lg = logging.getLogger('sha1p')

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(color_formatter)
logging.getLogger("").handlers = [] #leip defined handlers
lg.setLevel(logging.INFO)
lg.addHandler(stream_handler)

lg.info('start sha1p')


SHADATA = collections.defaultdict(list)
LOCKED = ""
JOBS_DONE = 0
SCHEDULED = 0

def write_both_checksums(dirname, files):
    lg.debug("writing sha1 checksum")
    write_to_checksum_file(
        os.path.join(dirname, 'SHA1SUMS'),
        [[x[0], x[1]] for x in files]
        )
    lg.debug("writing qd checksum")
    write_to_checksum_file(
        os.path.join(dirname, 'QDSUMS'),
        [[x[0], x[2]] for x in files]
        )

def write_to_checksum_file(hashfile, files):

    j = 0
    lg.debug("writing %d sha1sums to %s", len(files), hashfile)

    hashes = {}

    #read old sha1file
    if os.path.exists(hashfile):
        with open(hashfile) as F:
            for line in F:
                hsh, fn = line.strip().split(None, 1)
                hashes[fn] = hsh

    lg.debug("found %d hashes", len(hashes))

    #insert our sha1 - possibly overwriting other version
    for fn, hs in files:
        j += 1
        hashes[fn] = hs

    #write new sha1file
    lg.debug("now has %d hashes", len(hashes))

    try:
        lg.debug("start writing %d hashs to %s", len(hashes), hashfile)
        with open(hashfile, 'w') as F:
            for fn in sorted(hashes.keys()):
                if fn in ['QDSUMS', 'SHA1SUMS']:
                    continue
                F.write("{}  {}\n".format(hashes[fn], fn))

    except IOError:
        lg.warning("can not write to checksum file: %s", hashfile)
        return

    #fix permissions  - but only when root
    if os.geteuid() != 0:
        return j

    #change SHA1SUM file
    dirname = os.path.dirname(hashfile)
    if not dirname.strip():
        dirname = '.'
    dstats = os.stat(dirname)

    if os.path.exists(hashfile):
        os.chmod(hashfile, 0o664)
        os.chown(hashfile, dstats.st_uid, dstats.st_gid)

    return j



def process_file(*args, **kwargs):
    global JOBS_DONE
    try:
        process_file_2(*args, **kwargs)
        JOBS_DONE += 1
    except:
        import traceback
        traceback.print_exception()
        exit(-1)

def process_file_2(datalock, i, fn, force, echo):

    global SHADATA, LOCKED, JOBS_DONE, SCHEDULED

    filename = os.path.basename(fn)
    dirname = os.path.dirname(fn)

    qd_hash_file = os.path.join(dirname, 'QDSUMS')
    sha1_hash_file = os.path.join(dirname, 'SHA1SUMS')

    sha1_file = hash.check_hashfile(sha1_hash_file, filename)

    qd = hash.get_qdhash(fn)
    qd_file = hash.check_hashfile(qd_hash_file, filename)

    if (not force) and (not sha1_file is None) and (qd == qd_file):
        #nothing changed - sha1 is present - not force - return
        return

    #if not in the shasum files - then - maybe there is an old .mad file
    #
    sha1 = None

    madfile = os.path.join(dirname, '.{}.mad'.format(filename))
    if os.path.exists(madfile):
        with open(madfile) as F:
            maf = yaml.load(F)
            if 'hash' in maf:
                sha1 = maf['hash'].get('sha1')
                qd_file = maf['hash'].get('qdhash')

    if qd_file == qd and (not sha1 is None):
        #use mad file sha1
        lg.debug('reusing .mad hash: %s is %s', fn, sha1)
    else:

        sha1 = hash.get_sha1sum(fn)
        lg.debug('calculated hash: %s is %s', fn, sha1)

    datalock.acquire() #processing the SHADATA global data structure - lock
    assert(LOCKED == "")
    LOCKED = "YES"

    #lg.debug("updating SHADATA")

    SHADATA[dirname].append((filename, sha1, qd))

    if i > 0 and i % 100 == 0:
        for dirname in SHADATA:
            if len(SHADATA[dirname]) == 0:
                continue
            #lg.debug('flushing to dir: "%s"', dirname)
            write_both_checksums(dirname, SHADATA[dirname])
            SHADATA[dirname] = []
        lg.info('processed & written %d files (%d scheduled)',
                i, SCHEDULED)

    LOCKED = ""
    datalock.release()


def dispatch():

    global SCHEDULED

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--force', action='store_true')
    parser.add_argument('-d', '--do_dot_dirs', action='store_true')
    parser.add_argument('-j', '--threads', type=int, default=4)
    parser.add_argument('-e', '--echo', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-s', '--silent', action='store_true')
    parser.add_argument('file', nargs='*')

    args = parser.parse_args()

    if args.verbose:
        lg.setLevel(logging.DEBUG)
    if args.silent:
        lg.setLevel(logging.WARNING)

    pool = ThreadPool(args.threads)
    dlock = Lock()
    i = 0

    for i, fn in enumerate(util.get_filenames(args)):
        if '/.' in fn and (not args.do_dot_dirs):
            #no dot dirs
            lg.debug("ignoring in dotdir %s", fn)
            continue
        pool.apply_async(process_file, (dlock, i, fn, args.force, args.echo))
        SCHEDULED += 1

    lg.info(("processed all (%d) files - waiting for threads to " +
            "finish"),i)

    pool.close()

    pool.join()

    lg.info("finished - flushing cache")
    for dirname in SHADATA:
        lg.debug("flushing to %s/SHA1SUMS", dirname)
        if len(SHADATA[dirname]) == 0:
                continue
        write_both_checksums(dirname, SHADATA[dirname])



from __future__ import print_function

import re
import logging
import sys

import leip
from mad2.util import  get_filenames, get_all_mad_files, boolify

lg = logging.getLogger(__name__)

OPERATORS =  {'<' : lambda x,y: x < y,
              '<=' : lambda x,y: x <= y,
              '==' : lambda x,y: x == y,
              '=' : lambda x,y: x == y,
              '' : lambda x,y: x == y,
              '>' : lambda x,y: x > y,
              '>=' : lambda x,y: x > y,
              '!=' : lambda x,y: x != y,
              '!' : lambda x,y: x != y,
              '<>' : lambda x,y: x != y
              }


def is_numeric(val):
    if numre:
        return True
    return False

#@leip.arg('file', nargs='*')
@leip.arg('value', help='value to filter with',
    nargs='?', default=True)
@leip.arg('operator', help='operator (= < > != ! <>)',
    nargs='?', default='=')
@leip.arg('key', help='key to filter on')
@leip.command
def select(app, args):
    """
    Select a set of files based on their metadata

    This file needs to be fed filenames on stdin

    The command can take three forms

        mad find | mad select key operator value

        mad find | mad select key value # operator: =

        mad find | mad select key # only booleans, = True

    """
    val = args.value
    key = args.key
    oper = args.operator
    ofunc = OPERATORS.get(oper, lambda x,y: x==y)

    lg.debug("compare '{0}' '{1}' '{2}'".format(key, oper, val))

    def compare_func(what):
        try:
            if isinstance(what, int):
                return ofunc(what, int(val))
        except:
            pass
        try:
            if isinstance(what, float):
                return ofunc(what, float(val))
        except:
            pass
        try:
            if isinstance(what, bool):
                return ofunc(what, boolify(val))
        except:
            pass

        return ofunc(what, val)

    for madfile in get_all_mad_files(app, args):
        kv = madfile.mad.get(args.key, None)
        if kv is None:
            continue

        if compare_func(kv):
            print(madfile.filename)
#            print(madfile.fullpath)










from __future__ import print_function

import argparse
import logging
import os
import subprocess as sp
import shutil

import leip

from mad2.util import to_mad, get_mad_file

lg = logging.getLogger(__name__)

# @leip.arg('to')
# @leip.arg('fr', metavar='from')
# @leip.command
# def mmv(app, args):
#     """
#     Move a single file and the corresponding mad file.

#     If the file has already been moved - must move the
#     mad file.
#     """
#     fr = args.fr
#     frex = os.path.exists(fr)
#     frmad = get_mad_file(app, fr)
#     frmex = os.path.exists(frmad.madfile)

#     to = args.to
#     toex = os.path.exists(to)
#     tomad = get_mad_file(app, to)
#     tomex = os.path.exists(tomad.madfile)

#     if frex and toex:
#         lg.warning("Can't move %s", fr)
#         lg.warning(" | %s exists", to)
#         exit(-1)
# #    if frex:
# #        shutil.m

#     print("moving {} to {}".format(fr, to))

# #    if frex and not toex:
# #    if os.path.exists(fr) and not os


@leip.arg('args', nargs=argparse.REMAINDER)
@leip.command
def cp(app, args):
    """
    Fake copy - and make sure that the .mad files go along.
    """
    a = list(args.args)
    if not '-v' in a:
        a = ['-v'] + a
    a = ['cp'] + a
    P = sp.Popen(a, stdout=sp.PIPE)
    out, _ = P.communicate()
    froms, tos = [], []
    for line in out.split("\n"):
        line = line.strip()
        if not line: continue
        if line.count(' -> ') != 1:
            lg.warning("invalid cp output: ({}) {}".format(
                line.count(' -> '), line))
            continue
        fr, to = line.split(' -> ')
        froms.append(fr)
        tos.append(to)

    for fr, to in zip(froms, tos):
        frbase = fr.rsplit('/')[-1]
        if frbase[0] == '.' and frbase[-4:] == '.mad':
            #if this is a mad file, ignore
            continue

        frmad = to_mad(fr)
        tomad = to_mad(to)
        if frmad in froms:
            #if the mad file is already copied - don't bother
            continue
        shutil.copy(frmad, tomad)
        lg.debug("Copying mad file {} to {}".format(frmad, tomad))





from __future__ import print_function

import cPickle
import collections
import datetime
import logging
import os
import re
import time

import socket

import hashlib

import arrow
from pymongo import MongoClient
from termcolor import cprint
import yaml

import leip

from mad2.util import get_all_mad_files, humansize, persistent_cache
from mad2.ui import message

lg = logging.getLogger(__name__)

MONGO_SAVE_CACHE = []
MONGO_SAVE_COUNT = 0
MONGO_REMOVE_COUNT = 0
MONGO = None
MONGOCORE = None


def get_mongo_db(app):
    """
    Get the collection object
    """
    global MONGO

    if not MONGO is None:
        return MONGO

    mongo_info = app.conf['plugin.mongo']
    host = mongo_info.get('host', 'localhost')
    port = mongo_info.get('port', 27017)
    dbname = mongo_info.get('db', 'mad2')
    coll = mongo_info.get('collection', 'mad2')
    lg.debug("connect mongodb {}:{}".format(host, port))
    client = MongoClient(host, port)

    MONGO = client[dbname][coll]

    return MONGO


def get_mongo_core_db(app):
    """
    Get the core collection object
    """
    global MONGOCORE

    if not MONGOCORE is None:
        return MONGOCORE

    info = app.conf['store.mongo']
    host = info.get('host', 'localhost')
    port = info.get('port', 27017)
    dbname = info.get('db', 'mad2')
    coll = info.get('collection', 'dump')
    lg.debug("connect mongodb %s:%s/%s/%s", host, port, dbname, coll)
    client = MongoClient(host, port)

    MONGOCORE = client[dbname][coll]

    return MONGOCORE


def get_mongo_dump_id(mf):
    sha1sum = hashlib.sha1()
    sha1sum.update(mf['host'])
    sha1sum.update(mf['fullpath'])
    return sha1sum.hexdigest()[:24]


def mongo_prep_mad(mf):

    mongo_id = get_mongo_dump_id(mf)

    d = dict(mf)
    d['_id'] = mongo_id
    if 'uuid' in d:
        del d['uuid']
    if 'hash' in d:
        del d['hash']
    d['save_time'] = datetime.datetime.utcnow()

    return mongo_id, d

MONGO_SAVE_CACHE = []
MONGO_REMOVE_CACHE = []

def mongo_flush(app):

    global MONGO_REMOVE_CACHE
    global MONGO_SAVE_CACHE

    lg.debug("flush")
    if (len(MONGO_SAVE_CACHE) + len(MONGO_REMOVE_CACHE)) == 0:
        return

    collection = get_mongo_db(app)
    bulk = collection.initialize_unordered_bulk_op()

    for i, r in MONGO_SAVE_CACHE:
        bulk.find({'_id': i}).upsert().replace_one(r)
    res = bulk.execute()

    for i in MONGO_REMOVE_CACHE:
        #should try to do this in bulk, but uncertain how...
        lg.warning("removing id: %s", i)
        collection.remove({'_id': i})


    #print(res)
    lg.debug("Modified %d records", res['nModified'])
    MONGO_SAVE_CACHE = []
    MONGO_REMOVE_CACHE= []


def save_to_mongo(app, madfile):
    global MONGO_SAVE_COUNT
    global MONGO_SAVE_CACHE
    global MONGO_REMOVE_CACHE
    global MONGO_REMOVE_COUNT

    MONGO_SAVE_COUNT += 1

    mongo_id, newrec = mongo_prep_mad(madfile)

    if madfile['orphan']:
        MONGO_REMOVE_CACHE.append(mongo_id)
        lg.warning("removing %s from dump db", madfile['inputfile'])
    else:
        #print(newrec['host'])
        MONGO_SAVE_CACHE.append((mongo_id, newrec))

    lg.debug("prep for save: %s", madfile['inputfile'])
    if len(MONGO_SAVE_CACHE)  + len(MONGO_REMOVE_CACHE) > 33:
        mongo_flush(app)


@leip.hook("finish")
def save_to_mongo_finish(app):
    mongo_flush(app)

    #lg.critical('Finish')


@leip.hook("madfile_post_load")
def add_hook(app, madfile):
    madfile.mad['_id_dump'] = get_mongo_dump_id(madfile)


@leip.hook("madfile_save")
def store_in_mongodb(app, madfile):
    lg.debug("running store_in_mongodb")
    save_to_mongo(app, madfile)


@leip.subparser
def mongo(app, args):
    """
    Mongodb backend
    """
    pass  # this function is never called - it's just a placeholder


@leip.arg('file', nargs="*")
@leip.subcommand(mongo, "show")
def mongo_show(app, args):
    """
    Show mongodb records
    """
    MONGO = get_mongo_db(app)
    for madfile in get_all_mad_files(app, args):
        mongo_id = madfile['uuid']
        if mongo_id:
            print('#', mongo_id, madfile['filename'])
            rec = MONGO.find_one({'_id': mongo_id})
            # print(madfile.filename)
            if not rec:
                continue
            for key in rec:
                if key == '_id':
                    print('uuid\t{1}'.format(key, rec[key]))
                    continue
                print('{0}\t{1}'.format(key, rec[key]))


@leip.flag('-c', '--core')
@leip.arg('mongo_id')
@leip.subcommand(mongo, "get")
def mongo_get(app, args):
    """
    get a mongodb record based on id
    """
    MONGO_D = get_mongo_db(app)

    mongo_id = args.mongo_id

    rec = MONGO_D.find_one({'_id': mongo_id})

    if args.core:
         MONGO_C = get_mongo_core_db(app)
         core_id = rec['sha1sum'][:24]
         rec = MONGO_C.find_one({'_id': core_id})


    if not rec:
        return
    print(yaml.safe_dump(rec, default_flow_style=False))


@leip.flag('-c', '--core')
@leip.arg('mongo_id')
@leip.subcommand(mongo, "del")
def mongo_del(app, args):
    """
    get a mongodb record based on id
    """
    if args.core:
        MONGO = get_mongo_core_db(app)
    else:
        MONGO = get_mongo_db(app)

    mongo_id = args.mongo_id
    MONGO.remove({'_id': mongo_id})


@leip.subcommand(mongo, "count")
def mongo_count(app, args):
    """
    Show the associated mongodb record
    """
    MONGO_mad = get_mongo_db(app)
    print(MONGO_mad.count())


@leip.arg('-n', '--no', type=int, default=10)
@leip.subcommand(mongo, "last")
def mongo_last(app, args):
    now = arrow.now()
    MONGO_mad = get_mongo_db(app)
    res = MONGO_mad.aggregate([
        {"$sort" : { "save_time": -1 }},
        {"$limit" : args.no},
    ])
    for i, r in enumerate(res['result']):
        if i > args.no:
            break
        print("\t".join(
            [ arrow.get(r['save_time']).humanize(),
              r['filename'], r.get('_id', '')]))


@leip.flag('--delete')
@leip.arg('-u', '--username')
@leip.arg('-b', '--backup')
@leip.arg('-B', '--ignore_backup_volumes')
@leip.arg('-v', '--volume')
@leip.arg('-p', '--project')
@leip.arg('-P', '--pi')
@leip.arg('-e', '--experiment')
@leip.arg('-H', '--host')
@leip.arg('-s', '--sha1sum')
@leip.command
def search(app, args):
    """
    Find files
    """

    MONGO_mad = get_mongo_db(app)

    query = {}

    for f in ['username', 'backup', 'volume', 'host',
              'sha1sum', 'project', 'project', 'pi']:
        if not f in args:
            continue

        v = getattr(args, f)
        if v is None:
            continue
        query[f] = v

    if args.delete:
        MONGO_mad.remove(query)
        return

    res = MONGO_mad.find(query)
    for r in res:
        print(r['fullpath'])


@persistent_cache(leip.get_cache_dir('mad2', 'mongo', 'sum'),
                  'group_by', 60 * 60 * 24)
def _single_sum(app, group_by=None, force=False):
    groupby_field = "${}".format(group_by)
    MONGO_mad = get_mongo_db(app)

    res = MONGO_mad.aggregate([
        {"$match" : {"orphan" : False}},
        {'$group': {
            "_id": groupby_field,
            "total": {"$sum": "$filesize"},
            "count": {"$sum": 1}}},
        {"$sort" : { "total": -1
                  }}])

    return res['result']


@leip.flag('-f', '--force', help='force query (otherwise use cache, and'
           + ' query only once per day')
@leip.flag('-H', '--human', help='human readable')
@leip.arg('group_by', nargs='?', default='host')
@leip.subcommand(mongo, "sum")
def mongo_sum(app, args):
    """
    Show the associated mongodb record
    """

    res = _single_sum(app, group_by=args.group_by, force=args.force)
    total_size = long(0)
    total_count = 0

    mgn = len("Total")
    for reshost in res:
        gid = reshost['_id']
        if gid is None:
            mgn = max(4, mgn)
        else:
            mgn = max(len(str(reshost['_id'])), mgn)

    fms = "{:" + str(mgn) + "}\t{:>10}\t{:>9}"
    for reshost in res:
        total = reshost['total']
        count = reshost['count']
        total_size += long(total)
        total_count += count
        if args.human:
            total_human = humansize(total)
            print(fms.format(
                reshost['_id'], total_human, count))
        else:
            print("{}\t{}\t{}".format(
                reshost['_id'], total, count))

    if args.human:
        total_size_human = humansize(total_size)
        print(fms.format(
            "Total", total_size_human, count))
    else:
        print("Total\t{}\t{}".format(total_size, total_count))


@leip.flag('-H', '--human', help='human readable')
@leip.flag('-s', '--sort_on_field')
@leip.arg('group_by_2')
@leip.arg('group_by_1')
@leip.subcommand(mongo, "sum2")
def mongo_sum2(app, args):
    """
    Show the associated mongodb record
    """
    gb1_field = "${}".format(args.group_by_1)
    gb2_field = "${}".format(args.group_by_2)

    gb_pair_field = "${}_${}".format(gb1_field, gb2_field)

    MONGO_mad = get_mongo_db(app)

    if args.sort_on_field:
        sort_field = '_id'
        sort_order = 1
    else:
        sort_field = 'total'
        sort_order = -1
    res = MONGO_mad.aggregate([
            {"$match" : {"orphan" : False}},
            {'$group': {
                "_id": {
                    "group1": gb1_field,
                    "group2": gb2_field },
                "total": {"$sum": "$filesize"},
                "count": {"$sum": 1}}},
             {"$sort" : {
              "sort_field": sort_order
              }}
        ])
    total_size = 0
    total_count = 0

    gl1 = gl2 = len("Total")

    for r in res['result']:
        #print(r)
        g1 = str(r['_id'].get('group1'))
        g2 = str(r['_id'].get('group2'))
        gl1 = max(gl1, len(g1))
        gl2 = max(gl2, len(g2))

    fms = "{:" + str(gl1) + "}  {:" + str(gl2) + "}  {:>10}  {:>9}"
    for r in res['result']:
        g1 = str(r['_id'].get('group1', '-'))
        g2 = str(r['_id'].get('group2', '-'))
        total = r['total']
        count = r['count']
        total_size += total
        total_count += count
        if args.human:
            total = humansize(total)
            print(fms.format(g1,g2, total, count))
        else:
            print("{}\t{}\t{}\t{}".format(g1, g2, total, count))

    if args.human:
        total_size = humansize(total_size)
        print(fms.format(
            "Total", "", total, count))
    else:
        print("Total\t\t{}\t{}".format(total_size, total_count))



@persistent_cache(leip.get_cache_dir('mad2', 'mongo', 'sum'),
                  'name', 60 * 60 * 23)
def _complex_sum(app, name, fields=['username', 'host'],
                 force=False):

    MONGO_mad = get_mongo_db(app)
    qid = dict([(x, "$" + x) for x in fields])
    aggp = [{'$group': {
                "_id": qid,
                "total": {"$sum": "$filesize"},
                "count": {"$sum": 1}}}]
    res = MONGO_mad.aggregate(aggp)
    return res['result']

@leip.flag('-f', '--force', help='force query (otherwise use cache, and'
           + ' query only once per day')
@leip.arg('output_file', help='{stamp} will be replaced by a timestamp')
@leip.subcommand(mongo, "csum")
def complex_sum(app, args):
    """
    Query on a number of fields
    """

    fields = [
        'username',
        'experiment_type',
        'host',
        'volume',
        'orphan',
        'filetype',
        'organism',
        'biologist',
        'genome_build',
        'experiment_type',
        'pi',
        'project',
        'category',
        'backup']


    res = _complex_sum(app, name='table_query',
                       fields=fields, force=args.force)

    utc = arrow.now()
    stamp = utc.format('YYYY-MM-DD')

    data = collections.defaultdict(list)
    precords = []
    for i, rec in enumerate(res):
        _id = rec['_id']
        prec = {}
        prec['count'] = rec['count']
        prec['sum'] = rec['total']
        for f in fields:
            prec[f] = _id.get(f)
        precords.append(prec)

    out = args.output_file.format(stamp=stamp)
    with open(out, 'w') as F:
        F.write(yaml.safe_dump(precords, default_flow_style=False))


@leip.flag('--remove-from-core', help='also remove from the core db')
@leip.arg('file', nargs="*")
@leip.command
def forget(app, args):
    MONGO = get_mongo_db(app)
    MONGO_CORE = get_mongo_core_db(app)
    to_remove = []
    to_remove_core = []

    def go(coll, lst):
        coll.remove( {'_id' : { '$in' : lst } } )

    for madfile in get_all_mad_files(app, args):
        to_remove.append(madfile['_id_dump'])
        if args.remove_from_core:
            to_remove_core.append(['_id'])

        if len(to_remove) > 100:
            go(MONGO, to_remove)
            to_remove = []
        if len(to_remove_core) > 100:
            go(MONGO_CORE, to_remove_core)
            to_remove_core = []

    go(MONGO, to_remove)
    go(MONGO_CORE, to_remove_core)


@leip.flag('-r', '--remove-from-dump', help='remove from the dump db')
@leip.flag('--remove-from-core', help='also remove from the core db')
@leip.flag('--run', help='actually run - otherwise it\'s a dry run showing ' +
           'what would be deleted')
@leip.arg('dir', nargs="?", default='.')
@leip.flag('-e', '--echo', help='echo all files')
@leip.command
def forget_dir(app, args):
    args.forget_all = True
    flush_dir(app, args)


@leip.flag('-r', '--remove-from-dump', help='remove from the dump db '
           + 'otherwise files in dump are tagged as deleted')
@leip.flag('--remove-from-core', help='also remove from the core db')
@leip.flag('--run', help='actually run - otherwise it\'s a dry run showing ' +
           'what would be deleted')
@leip.flag('--forget-all', help='forget all files in this directory ' +
           'and below')
@leip.flag('-e', '--echo', help='echo all files')
@leip.arg('dir', nargs='?', default='.')
@leip.command
def flush_dir(app, args):
    """
    Recursively flush deleted files from the dump db
    """

    if args.remove_from_core and not args.remove_from_dump:
        lg.warning("This is an odd command to give - why remove from the core")
        lg.warning("  and not the dump db?")
        exit()

    MONGO_mad = get_mongo_db(app)
    MONGO_core = get_mongo_core_db(app)

    host = socket.gethostname()
    wd = os.path.abspath(os.getcwd())

    rex = re.compile("^" + wd)

    query = {
        'host' : host,
        'dirname' : rex,
        'orphan': False}

    ids_to_remove = []
    core_to_remove = []

    if args.run:
        pass

    res = MONGO_mad.find(query)
    for r in res:

        if os.path.exists(r['fullpath']):
            if args.echo:
                print("+ " + r['fullpath'])
            if not args.forget_all:
                continue
        else:
            print("- " + r['fullpath'])

        ids_to_remove.append(r['_id'])

        if args.remove_from_core:
            core_to_remove.append(r['sha1sum'][:24])

        if args.run and len(ids_to_remove) >= 100:
            if args.remove_from_dump:
                lg.warning("removing %d records", len(ids_to_remove))
                MONGO_mad.remove( {'_id' : { '$in' : ids_to_remove } } )
            else:
                lg.warning("setting %d records to orphaned", len(ids_to_remove))

            ids_to_remove = []

        if args.run and args.remove_from_core and \
                len(core_to_remove) >= 200:
            lg.warning("removing %d records from core",
                       len(core_to_remove))
            MONGO_core.remove( {'_id' : { '$in' : core_to_remove } } )
            core_to_remove = []

    if args.run:
        if args.remove_from_dump:
            lg.warning("removing %d records", len(ids_to_remove))
            MONGO_mad.remove( {'_id' : { '$in' : ids_to_remove } } )
        else:
            lg.warning("setting %d records to orphaned", len(ids_to_remove))
            print(ids_to_remove)
            MONGO_mad.update(
                {'_id' : { '$in' : ids_to_remove } },
                {'$set': {'orphan': True}}
                )

    if args.run and args.remove_from_core:
            lg.warning("removing %d records from core",
                       len(core_to_remove))
            MONGO_core.remove( {'_id' : { '$in' : core_to_remove } } )

@leip.flag('-f', '--force')
@leip.subcommand(mongo, "drop")
def mongo_drop(app, args):
    """
    Show the associated mongodb record
    """
    if not args.force:
        print("use --force to really drop the database")
        exit()

    MONGO_mad = get_mongo_db(app)
    MONGO_mad.drop()


@leip.flag('-e', '--echo')
@leip.arg('file', nargs="*")
@leip.subcommand(mongo, "save")
def mongo_save(app, args):
    """
    Save to mongodb
    """
    for madfile in get_all_mad_files(app, args):
        lg.debug("save to mongodb: %s", madfile['inputfile'])
        save_to_mongo(app, madfile)
        if args.echo:
            print(madfile['inputfile'])


@leip.subcommand(mongo, "create_index")
def mongo_index(app, args):
    """
    Ensure indexes on the relevant fields
    """
    MONGO_mad = get_mongo_db(app)
    MONGO_core = get_mongo_core_db(app)

    for f in app.conf['plugin.mongo.indici']:
        print("create index on: {}".format(f))
        MONGO_mad.ensure_index(f)
        MONGO_core.ensure_index(f)


@persistent_cache(leip.get_cache_dir('mad2', 'mongo', 'keys'),
                  1, 60*60*24)
def _get_mongo_keys(app, collection, force=False):
    from bson.code import Code
    mapper = Code("""
        function() {
            for (var key in this) { emit(key, 1); }
        } """)

    rv = {}

    reducer = Code("function(key, vals) { return Array.sum(vals); }")
    if collection == 'dump':
        message("Get keys from the dump db")
        COLLECTION = get_mongo_db(app)
    else:
        message("Get keys from the core db")
        COLLECTION = get_mongo_core_db(app)

    res =  COLLECTION.map_reduce(mapper, reducer, "my_collection" + "_keys")

    for r in res.find():
        rv[r['_id']] = int(r['value'])

    return rv

@leip.flag('-f', '--force')
@leip.arg('output_file', help='{stamp} will be replaced by a timestamp')
@leip.subcommand(mongo, 'keys')
def mongo_keys(app, args):
    dump = _get_mongo_keys(app, 'dump', force=args.force)
    core = _get_mongo_keys(app, 'core', force=args.force)
    all_keys = list(sorted(set(dump.keys()) | set(core.keys())))

    max_key_len = max([len(x) for x in all_keys])

    utc = arrow.now()
    stamp = utc.format('YYYY-MM-DD')

    print( ('%-' + str(max_key_len) + 's: %12s %12s') % (
          '#key', 'dump', 'core'))

    records = []

    for k in all_keys:
        records.append({
            'key': k,
            'dump' : dump.get(k),
            'core': core.get(k) })

        print( ('%-' + str(max_key_len) + 's: %12s %12s') % (
              k, dump.get(k, ''), core.get(k, '')))

    out = args.output_file.format(stamp=stamp)
    with open(out, 'w') as F:
        F.write(yaml.safe_dump(records, default_flow_style=False))


@leip.arg('keyname', metavar='key')
@leip.command
def distinct(app, args):
    dump = get_mongo_db(app)
    vals = sorted(dump.distinct(args.keyname))
    if None in vals:
        cprint('<None>', 'yellow')
    for v in vals:
        if v is None:
            continue
        cprint('"' + v + '"', 'green')


@leip.flag('--I-am-sure')
@leip.arg('key')
@leip.arg('database', choices=['core', 'dump'])
@leip.subcommand(mongo, 'remove_key_from_db')
def mongo_remove_key(app, args):
    lg.warning("removing %s from the %s db", args.key, args.database)
    if args.database == 'core':
        COLLECTION = get_mongo_core_db(app)
    elif args.databse == 'dump':
        COLLECTION = get_mongo_db(app)

    print(COLLECTION)
    query = {args.key : {'$exists': True}}
    update = {"$unset": { args.key: "" }}
    print(query)
    print(update)
    COLLECTION.update(query, update, multi=True)


@persistent_cache(leip.get_cache_dir('mad2', 'mongo', 'command'),
                  1,  1)
def _run_mongo_command(app, name, collection, query, kwargs={}, force=False):
    """
    Execute mongo command.
    """
    MONGO_mad = get_mongo_db(app)
    res = MONGO_mad.database.command(query, collection, **kwargs)
    return res



WASTE_PIPELINE = [
    {"$match" : {"orphan" : False}},
    {"$sort": {"sha1sum": 1 }},
    {"$project": {"filesize": 1,
                  "sha1sum": 1,
                  "usage": {"$divide": ["$filesize", "$nlink"]}}},
    {"$group": {"_id": "$sha1sum",
                 "no_records": {"$sum": 1 },
                 "mean_usage": {"$avg": "$usage" },
                 "total_usage": {"$sum": "$usage" },
                 "filesize": {"$max": "$filesize" }}},
     {"$project": {"filesize": 1,
                   "sha1sum": 1,
                   "total_usage": 1,
                   "waste": {"$subtract": ["$total_usage", "$filesize"]}}},
    {"$match": {"waste": {"$gt": 500}}},
    {"$group": {"_id": None,
                "no_files": {"$sum": 1 },
                "waste": {"$sum": "$waste" }}}]


@leip.flag('-e', '--echo')
@leip.arg('file', nargs="*")
@leip.command
def repl(app, args):
    """
    Save to mongodb
    """

    MONGO_mad = get_mongo_db(app)

    backup_hosts = set()
    for host in app.conf['host']:
        if app.conf['host'][host]['backup']:
            backup_hosts.add(host)


    check_shasums = False
    if args.file > 0:
        for f in args.file:
            if os.path.exists(f): break
            if len(f) != 40: break
        else: check_shasums = True

    def _process_query(query):
        res = MONGO_mad.find(query)
        for r in res:
            days = (arrow.now() - arrow.get(r['save_time'])).days
            symlink = r.get('is_symlink', False)
            if symlink:
                stag = 'S'
            else:
                stag = '.'
            cprint('%1d%s' % (r['nlink'], stag), 'yellow', end=" ")
            cprint('%3d' % days, 'green', end="d ")
            cprint('%6s' % humansize(r['filesize']), 'white', end=" ")
            if r['host'] in backup_hosts:
                cprint(r['host'], 'green', attrs=['bold'], end=':')
            else:
                cprint(r['host'], 'cyan', end=':')
            cprint(r['fullpath'])

    if check_shasums:
        for sha1sum in args.file:
            query = {'sha1sum' : sha1sum}
            _process_query(query)
    else:
        for madfile in get_all_mad_files(app, args):
            #print(madfile['sha1sum'])
            query = {'sha1sum': madfile['sha1sum']}
            _process_query(query)



FIND_WASTER_PIPELINE = [
    {"$match" : {"orphan" : False}},
    {"$project": {"filesize": 1,
                  "sha1sum": 1,
                  "usage": {"$divide": ["$filesize", "$nlink"]}}},
    {"$group": {"_id": "$sha1sum",
                "no_records": {"$sum": 1 },
                "mean_usage": {"$avg": "$usage" },
                "total_usage": {"$sum": "$usage" },
                "filesize": {"$max": "$filesize" }}},
    {"$project": {"filesize": 1,
                  "total_usage": 1,
                  "waste": {"$subtract": ["$total_usage", "$filesize"]}}},
    {"$match": {"waste": {"$gt": 500}}},
    {"$sort" : {"waste": -1} },
    {"$limit": 100}]


@persistent_cache(leip.get_cache_dir('mad2', 'mongo', 'waste'), 1,  24*60*60)
def _run_waste_command(app, name, force=False):
    """
    Execute mongo command.
    """
    MONGO_mad = get_mongo_db(app)
    res = MONGO_mad.aggregate(FIND_WASTER_PIPELINE, allowDiskUse=True)
    return res


@leip.flag('-N', '--no-color', help='no ansi coloring of output')
@leip.arg('-n', '--no-records', default=20, type=int)
@leip.flag('-f', '--force')
@leip.command
def waste(app, args):

    db = get_mongo_db(app)

    res = _run_waste_command(app, 'waste_pipeline',
                             force=args.force)['result']

    for i, r in enumerate(res):
        if i >= args.no_records:
            break

        sha1sum = r['_id']
        if not sha1sum.strip():
            continue

        cprint(sha1sum, 'grey', end='')
        cprint(" sz ", "grey", end="")
        cprint("{:>9}".format(humansize(r['waste'])), end='')
        cprint(" w ", "grey", end="")
        cprint("{:>9}".format(humansize(r['filesize'])),end='')

        hostcount = collections.defaultdict(lambda: 0)
        hostsize = collections.defaultdict(lambda: 0)
        owners = set()
        for f in db.find({'sha1sum': sha1sum}):
            owners.add(f['username'])
            host = f['host']
            hostcount[host] += 1
            hostsize[host] += float(f['filesize']) / float(f['nlink'])

        for h in hostcount:
            print(' ', end='')
            cprint(h, 'green', end=':')
            cprint(hostcount[h], 'cyan', end="")

        cprint(" ", end="")
        cprint(", ".join(owners), 'red')


@leip.arg('-S', '--subject')
@leip.flag('-f', '--force')
@leip.command
def waste_text_report(app, args):

    db = get_mongo_db(app)

    res = _run_waste_command(app, 'waste_pipeline',
                             force=args.force)['result']


    if args.subject:
        print("Subject: {}".format(args.subject))


    #This week's winner
    top = res[0]
    sha1sum = top['_id']
    owners = set()
    hostcount = collections.defaultdict(lambda: 0)
    hostsize = collections.defaultdict(lambda: 0)


    total = 0
    for rec in db.find({'sha1sum': sha1sum}):
        total += 1
        host = rec['host']
        hostcount[host] += 1
        hostsize[host] += float(rec['filesize']) / float(rec['nlink'])
        owners.add(rec['username'])



    print("This week's winner: {}".format(", ".join(owners)))
    print("One file, ", end="")
    print("{} location".format(total), end="")
    if total > 1: print("s", end="")
    print(", {} server,".format(len(hostcount)), end="")
    if len(hostcount) > 1: print("s", end="")
    print(" wasting {}.".format(humansize(top['waste'])))
    print("try:\n   mad repl {}\n\n".format(sha1sum))

    no_to_print = 20
    print("Waste overview: (no / sha1sum / waste / filesize)")
    print("=================================================\n")
    for i, r in enumerate(res):
        if i >= no_to_print:
            break

        sha1sum = r['_id']
        if not sha1sum.strip():
            continue
        print("{:2d} {} {:>10} {:>10}"\
                    .format(i, sha1sum, humansize(r['waste']),
                            humansize(r['filesize'])))


    print("\n\nDetails: (nlink/symlink/size/owner)")
    print("===================================")
    import textwrap
    for i, r in enumerate(res):
        if i >= no_to_print:
            break

        sha1sum = r['_id']
        if not sha1sum.strip():
            continue

        print("# {:2d} {} {:>10} {:>10}"\
                    .format(i, sha1sum, humansize(r['waste']),
                            humansize(r['filesize'])))

        records = collections.defaultdict(list)
        hostcount = collections.defaultdict(lambda: 0)
        hostsize = collections.defaultdict(lambda: 0)

        for rec in db.find({'sha1sum': sha1sum}):
            host = rec['host']
            records[host].append(rec)
            hostcount[host] += 1
            hostsize[host] += float(rec['filesize']) / float(rec['nlink'])


        for h in hostcount:
            print("## Host: {}, copies: {}, total use: {}".format(
                      h, hostcount[h], humansize(hostsize[h])))
            for rec in records[host]:
                smarker = '.'
                if rec.get('is_symlink'):
                    smarker = 'S'
                print("  {} {}".format(rec.get('nlink', '?'), smarker),
                      end=' ')
                print(humansize(rec['filesize']), end=' ')
                print(rec['username'])
                print("   " + rec['fullpath'])
#                for j, pp in enumerate(textwrap.wrap(rec['fullpath'], 70)):
#                    print(" " * 8 + pp)


        print("")




import os
import logging
import subprocess
import sys
import tempfile
import glob


def message(txt, *args):
    """
    Message to the user (always on stderr)
    """
    if not sys.stderr.isatty():
        sys.stderr.write('Mad: ' + txt.rstrip() % args + "\n")
        return

    sys.stderr.write('\033[1;97;48;5;70mMad\033[0m: ')
    sys.stderr.write(txt.strip() % args + "\n")

def error(txt, *args):
    """
    Message to the user (always on stderr)
    """
    if not sys.stdout.isatty():
        sys.stderr.write("Error: " + txt.strip() % args + "\n")
        return
    sys.stderr.write('\033[1;97;48;5;124mError:\033[0m')
    sys.stderr.write(txt.strip() % args + "\n")

def errorexit(txt, exit_code=1):
    error(txt)
    exit(exit_code)


# Put site-packages in front of sys.path so we don't end
# up importing the global readline.so - osx hack
sys.path = (
    [p for p in sys.path if 'site-packages' in p] + \
    [p for p in sys.path if 'site-packages' not in p])

# hack - otherwise readline outputs unwanted control
# character
if os.environ.has_key('TERM'):
    if os.environ['TERM'] == 'xterm':
        os.environ['TERM'] = 'vt100'

import readline

lg =logging.getLogger(__name__)



###############################################################################
##
## readline enabled user prompt
##
#########################################################################

## See if we can do intelligent things with job variables
def untangle(txt):
    print("untangle", txt)
    return txt
    #return sysConf.job.conf.interpret(txt)

## Handle mad directories
_FSCOMPLETECACHE = {}

def fsCompleter(text, state):
    def g(*a):
        pass
        #with open('/tmp/fscomp.%d.log' % os.getuid(), 'a') as F:
        #    F.write("\t".join(map(str, a)) + "\n")

    g("text   : ", text)
    g("state  :", state)

    if _FSCOMPLETECACHE and text in _FSCOMPLETECACHE.keys():
        try:
            #rv = _FSCOMPLETECACHE[text]
            #g(str(rv))
            rv = _FSCOMPLETECACHE[text][state]
            g('from cache', rv)
            return rv
        except:
            g('%s' % _FSCOMPLETECACHE)
            g('cache problem')
            import traceback
            E = traceback.format_exc()
            g(E)
            return None

    detangle = False
    utext = text
    #see if there are templates in the text - if so - untangle
    if '{{' in text or '{%' in text:
        #do a complete untangle first
        utext = untangle(text)
        g("untang  :", utext)
        detangle = True
        #if utext != text:
        #    detangle = True
    else:
        g("no detangle")

    g("utext : " + utext)

    #find the last word - to expand
    #string: stored in 'ctext'. The rest is in 'prefix'
    if ' ' in utext:
        addPrefix = True
        prefix, uptext = utext.rsplit(' ', 1)
    else:
        addPrefix = False
        prefix, uptext = "", utext

    g('prefix :', prefix)
    g('uptext :', uptext)

    if os.path.isdir(uptext) and not uptext[-1] == '/':
        sep = '/'
    else: sep = ''


    if prefix or uptext[:2] == './' or \
            uptext[:3] == '../' or uptext[0] == '/':
        #try to expand path
        #get all possibilities
        pos = glob.glob(uptext + sep + '*')
    else:
        pos = []

    np = []
    for i, p in enumerate(pos):
        g("found %s" % p)
        if os.path.isdir(p) and not p[-1] == '/':
            p += '/'
        if addPrefix:
            p = prefix + ' ' + p
        if detangle:
            g("detangling")
            g("from %s" % p)
            g("replacing %s" % utext)
            g("with %s" % text)
            p = p.replace(utext , text)
        np.append(p)
    g('pos', np)

    _FSCOMPLETECACHE[text] = np
    g(_FSCOMPLETECACHE)
    try:
        rv = np[state]
        return rv
    except IndexError:
        return None


def _get_history_file(appname, category):
    """
    return the history file for this parameter
    """

    histdir =  os.path.join(
            os.path.expanduser('~'), '.config', appname,
                               'history')

    if not os.path.exists(histdir):
        os.makedirs(histdir)

    histfile = os.path.join(histdir, category)
    return histfile

def _check_history_duplicates(value):
    # prevent duplicates in histtory the last ten lines
    # of the history - if the item is already there -
    # remove it again
    histlen = readline.get_current_history_length()
    if histlen > 1:
        for i in range(max(histlen-10, 0), histlen):
            previtem = readline.get_history_item(i)
            if previtem == value:
                lg.debug("removing duplicate %s" % value)
                readline.remove_history_item(histlen-1)
                break

def askUserEditor(default):
    editor = os.environ.get('EDITOR', 'vim')
    tmp_file = tempfile.NamedTemporaryFile('wb', delete=False)

    #write default value to the tmp file
    if default:
        tmp_file.write(default + "\n")
    else:
        tmp_file.write("\n")
    tmp_file.close()

    tty = open('/dev/tty')

    subprocess.call('{} {}'.format(editor, tmp_file.name),
        stdin=tty, shell=True)
    sys.stdin = sys.__stdin__


    #read value back in
    with open(tmp_file.name, 'r') as F:
        #removing trailing space
        val = F.read().rstrip()
    #remove tmp file
    os.unlink(tmp_file.name)

def askUser(parameter, appname='mad', default="", data = {},
            xtra_history=None, prompt=None):
    """
    :param parameter: paramter to ask value of
    :param default: default value - if absent use the last
      history item
    :param xtra_history: extra history file to show to the user
    """

    if not default:
        default = ""
    lg.debug("askUser {0} ({1})".format(parameter, default))

    def startup_hook():
        lg.debug("readline starup hook ({0})".format(default))
        readline.insert_text(str(default))
        #readline.redisplay()

    if xtra_history:
        history_file = xtra_history
    else:
        history_file = _get_history_file(appname, parameter)

    lg.debug("reading history from %s" % history_file)
    readline.clear_history()
    readline.set_completer_delims("\n`~!@#$^&*()-=+[]\|,?")
    readline.set_startup_hook(startup_hook)

    readline.set_completer(fsCompleter)
    readline.parse_and_bind("tab: complete")

    if history_file and os.path.exists(history_file):
        readline.read_history_file(history_file)

    hislen = readline.get_current_history_length()
    if default == '__last__':
        default = readline.get_history_item(hislen)

    if prompt is None:
        prompt = '%s: ' % parameter

    try:
        vl = raw_input(prompt)
    finally:
        readline.set_startup_hook()

    if not xtra_history:
        # if we're not reading from an xtra history file -
        # save it to the parameter history
        _check_history_duplicates(vl)
        readline.write_history_file(history_file)
    else:
        #see if we want to write to an additional history file
        lg.debug("xtra history file processed - now do boring one")
        history_file = _get_history_file(appname, parameter)
        lg.debug("boring read %s" % history_file)
        readline.clear_history()
        if os.path.exists(history_file):
            readline.read_history_file(history_file)
        readline.add_history(vl)
        _check_history_duplicates(vl)

        readline.write_history_file(history_file)

    return vl




from __future__ import print_function

import datetime
import logging
import sys
import os
import leip
import hashlib

from lockfile import FileLock

from mad2.util import get_all_mad_files
from mad2.hash import get_qdhash

lg = logging.getLogger(__name__)


def check_qdsumfile(qdfile, filename):
    lg.warning("DEPRECATED")

    if not os.path.exists(qdfile):
        return None
    with open(qdfile) as F:
        for line in F:
            hsh, fn = line.strip().split(None, 1)
            if fn == filename:
                return hsh
    return None


def append_qdsumfile(qdfile, filename, qd):
    lg.warning("DEPRECATED")

    qds = {}

    with FileLock(qdfile):
        #read old qdfile
        if os.path.exists(qdfile):
            with open(qdfile) as F:
                for line in F:
                    hsh, fn = line.strip().split(None, 1)
                    qds[fn] = hsh

        #insert our qd - possibly overwriting other version
        qds[filename] = qd

        #write new qdfile
        with open(qdfile, 'w') as F:
            for fn in sorted(qds.keys()):
                if fn in ['SHA1SUMS', 'QDSUMS']:
                    continue
                F.write("{}  {}\n".format(qds[fn], fn))


#@leip.hook("madfile_post_load", 250)
def qdhook(app, madfile):
    lg.warning("DEPRECATED")

    if madfile.get('orphan', False):
        # won't deal with orphaned files
        return

    if madfile.get('isdir', False):
        # won't deal with dirs
        return

    dirname = madfile['dirname']
    filename = madfile['filename']

    if filename in ['QDSUMS', 'SHA1SUMS']:
        return

    qdfile = os.path.join(dirname, 'QDSUMS')

    try:
        qd_file = check_qdsumfile(qdfile, filename)
    except:
        lg.warning("Error checking qdsum of %s in %s", filename, qdfile)

    qd = get_qdhash(madfile['fullpath'])

    if qd_file is None:
        #qd does not exists yet - create & return
        append_qdsumfile(qdfile, filename, qd)
        madfile.all['qdhash'] = qd
        return

    #else - check if the qd has chenged
    if qd_file == qd:
        #no? All is well -
        return

    lg.warning("'%s' may have hanged (qd hash did)", madfile['inputfile'])
    lg.warning("  - from file  : %s", qd_file)
    lg.warning("  - calculated : %s", qd)




import os
import sys

from mad2 import hash

def dispatch():

    if sys.argv[1] == '-c':
        sumfile = sys.argv[2]
        with open(sumfile) as F:
            for line in F:
                qd, fn = line.strip().split(None, 1)
                fn = fn.strip()
                if not os.path.exists(fn):
                    print "{}: FAILED no such file".format(fn)
                    continue

                qdnew = hash.get_qdhash(fn)
                if qd == qdnew:
                    print "{}: OK".format(fn)
                else:
                    print "{}: FAILED".format(fn)
        exit(0)

    for fn in sys.argv[1:]:
        if os.path.isdir(fn):
            sys.stderr.write("qdsum: {}: Is a directory\n".format(fn))
            continue
        if not os.path.exists(fn):
            sys.stderr.write("qdsum: {}: Does not exist\n".format(fn))
            continue

        qd = hash.get_qdhash(fn)
        print "{}  {}".format(qd, fn)




#!/usr/bin/env python

from setuptools import setup, find_packages

DESCRIPTION = """
file metadata tagger
"""

entry_points = {
    'console_scripts': [
        'mad = mad2.cli.main:dispatch',
        'sha1p = mad2.cli.sha1p:dispatch',
        'qdsum = mad2.cli.qdsum:dispatch'
    ]}

setup(name='mad2',
      version='0.1.17',
      description=DESCRIPTION,
      author='Mark Fiers',
      author_email='mark.fiers42@gmail.com',
      entry_points=entry_points,
      include_package_data=True,
      url='http://mfiers.github.com/mad2',
      packages=find_packages(),
      install_requires=[
          'Leip',
          'Fantail',
          'xlrd',
          'Jinja2',
          'lockfile',
          'python-dateutil',
          'pymongo',
          'colorlog',
          'arrow',
          'termcolor',
          'readline'
          ],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          ]
      )



from __future__ import print_function

import logging

import leip
from mad2.util import get_all_mad_files

lg = logging.getLogger(__name__)


@leip.arg('file', nargs='*')
@leip.arg('key', help='key to filter on')
@leip.commandName('print')
def mad_print(app, args):
    """
    print a single value from a single file
    """
    for madfile in get_all_mad_files(app, args):
        print(madfile.render(madfile[args.key], [madfile, app.conf]))




from __future__ import print_function

import collections
import logging
from multiprocessing.dummy import Pool as ThreadPool
import os
import subprocess as sp
import sys
import uuid

import leip
import fantail
from mad2.util import get_all_mad_files
from mad2.recrender import recrender

lg = logging.getLogger(__name__)
#lg.setLevel(logging.DEBUG)


def get_template(conf, filetype, name):
    """
    get a template branch based on filetype
    """
    if not filetype is None:
        tloc = 'template.{}.{}'.format(filetype, name)
        if tloc in conf:
            return conf[tloc]

    tloc = 'template.default.{}'.format(name)
    return conf.get(tloc)


@leip.flag('-e', '--exit_error', help='exit on no template found')
@leip.arg('-g', '--group', help='group on', default=1)
@leip.arg('-t', '--group_template', help='group template')
@leip.arg('file', nargs='*')
@leip.arg('name', help='template name to render')
@leip.command
def render(app, args):
    """
    Render a template

    Render a template in the context of a single madfile.

    ---
    A template is a configuration subtree that has the
    following keys:
        - template
        - defaults (optional)

    templates are to be found at the following locations in the
    configuration

       - template.{{filetype}}.{{template_name}}
       - template.default.{{template_name}}

    """

    lg.debug("rendering key %s", args.name)


    groupby = args.group
    try:
        groupby = int(groupby)
        numeric_groups = True
    except:
        numeric_groups = False

    mf_generator = get_all_mad_files(app, args)

    if numeric_groups:
        for res in render_numeric(app, mf_generator, args.name, groupby,
                                args.exit_error):
            print(res)

        return

    lg.warning("not implemented yet (-g %s)", group)



def render_numeric(app,
                   mf_generator,
                   name, count,
                   exit_error = False):

    elements = []

    group_template = app.conf['plugin.template.group_template']
    group_count = 0
    group_defaults = {}
    for madfile in mf_generator:

        filetype = madfile.get('filetype', None)
        template = get_template(app.conf, filetype, name)

        if template is None:
            if exit_error:
                lg.warning("No template '%s' for file '%s'",
                           name, madfile['inputfile'])
                exit()
            else:
                continue


        if 'group' in template:
            if group_template is None:
                group_template = template['group']
            else:
                assert(group_template == template['group'])

        group_defaults.update(template['defaults'])

        #render an element
        elements.append(madfile.render(
                            template['template'],
                            [template['defaults'],
                            app.conf]).strip())


        if len(elements) == count:
            if group_template is None:
                group_template = app.conf['plugin.template.group_template']
            #render a group
            result = recrender(group_template,
                    [{'elements': elements,
                      'group_count': group_count},
                      group_defaults,
                      app.conf])
            yield result
            elements = []


    if len(elements) > 0:
        assert(not group_template is None)

        #render a group
        result = recrender(group_template,
                [{'elements': elements,
                  'group_count': group_count},
                  group_defaults,
                  app.conf])
        yield result




import logging

import leip
import yaml


import mad2.util
from mad2.util import get_all_mad_files, get_mad_dummy
import mad2.hash

lg = logging.getLogger(__name__)

# make sure stores are cleaned up.
# note that initialization takes place in mad2.util - and only
# when a record is actually saved.
@leip.hook('finish')
def cleanup_stores(app):
    lg.debug("cleanup stores")
    mad2.util.cleanup_stores(app)


@leip.arg('file', nargs='*')
@leip.command
def echo(app, args):
    """
    Echo the filename

    note - this ensures that the sha1sum is calculated
    """
    for madfile in get_all_mad_files(app, args):
        print(madfile['inputfile'])


@leip.arg('-p', '--progress', action='store_true',
          help='show indication of progress')
@leip.flag('-e', '--echo')
@leip.arg('file', nargs='*')
@leip.command
def save(app, args):
    """
    save the file to the mad datastore

    note - this ensures that the sha1sum is calculated
    """
    lg.debug("start mad save")
    counter = 0
    for madfile in get_all_mad_files(app, args):
        lg.debug("processing %s", madfile['fullpath'])

        if madfile['orphan']:
            lg.warning("removing %s", madfile['inputfile'])
            lg.warning("sha1sum is/was: %s", madfile['sha1sum'])

        counter += 1
        if args.progress and counter % 1000 == 0:
            lg.warning("mad save: saved {} filesdif".format(counter))
        madfile.save()
        if args.echo:
            print madfile['inputfile']



def _save_dumped_doc(app, doc):
    dm = get_mad_dummy(app, doc)
    dm.save()

@leip.arg('dump_file', help='yaml dump file to load')
@leip.command
def load_dump(app, args):
    with open(args.dump_file) as F:
        for doc in yaml.load_all(F):
            _save_dumped_doc(app, doc)

@leip.arg('-o', '--output_file', help='outputf file to dump to',
          default='dump.yaml')
@leip.arg('file', nargs='*')
@leip.command
def dump(app, args):
    """
    dump a yaml representation of all files.
    """
    F = open(args.output_file, 'w')

    lg.debug("start mad dump")
    i = 0
    for madfile in get_all_mad_files(app, args):
        lg.debug("processing %s", madfile['fullpath'])

        if madfile['orphan']:
            continue

        if i > 0:
            F.write('---\n')

        F.write(madfile.pretty())
        i += 1

    F.close()
